# Script that receive from serial port the voltageChangeIntervalxDO value
# from the FPGA and calculate the correspective current.
# The communication format is the following:
# 1B of header (0xAB) + 5B of data (coming from LSB to MSB)
# Of the 40 bits of data, the first 39 are used. The rest is just zero padding.

import typer
import serial
import serial.tools.list_ports
from typing_extensions import Annotated
import datetime
from typing import Optional

from rich.console import Console
from rich.live import Live
from rich.text import Text
import time

app = typer.Typer(
    help="Command Line Interface for the ACCURATE2 evaluation board.",
    epilog="",
    add_completion=False
)

default_period = 100 # ms
default_lsb = 39.339 # aC
default_log_file = "output.log"
DAC_address = {'A': 0x00, 'B': 0x01, 'C': 0x02, 'D': 0x03, 'E': 0x04, 'F': 0x05, 'G': 0x06, 'H': 0x07}


@app.command()
def list_ports():
    '''
    List all the serial ports currently available.
    '''

    ports = serial.tools.list_ports.comports()
    for port in ports:
        print(port.device)


@app.command()
def open_serial(
    port: Annotated[
        str, typer.Argument(
            help="Serial port where the FPGA is connected.",
            envvar="SERIAL_PORT"
        )
    ],
    baudrate: Annotated[
        int, typer.Option(
            "-b", "--baudrate",
            help="Baudrate for the serial communication."
        )
    ] = 9600
):
    '''
    Open a serial connection with the FPGA. \n
    PORT accepts also the environment variable SERIAL_PORT.
    '''

    with serial.Serial(port, baudrate, timeout=1) as ser:
        print(f"Serial port {port} opened at {baudrate} baudrate.")
        while True:
            try:
                data = ser.read(1)
                print(data, end=' ', flush=True)
            except KeyboardInterrupt:
                print(f"\nSerial port {port} closed.")
                raise typer.Exit()


@app.command()
def get_current(
    port: Annotated[
        str, typer.Argument(
            help="Serial port where the FPGA is connected.",
            envvar="SERIAL_PORT"
        )
    ],
    baudrate: Annotated[
        int, typer.Option(
            "-b", "--baudrate",
            help="Baudrate for the serial communication."
        )
    ] = 9600,
    period: Annotated[
        int, typer.Option(
            "--period",
            help="Sampling period in ms."
        )
    ] = default_period,
    lsb: Annotated[
        float, typer.Option(
            "--lsb",
            help="Least significant bit value in aC."
        )
    ] = default_lsb,
    log: Annotated[
        Optional[str], typer.Option(
            "-l", "--log",
            help="Log the values to specified file.",
            show_default=False
        )
    ] = None,
    verbose: Annotated[
        bool, typer.Option(
            "-v", "--verbose",
            help="Enable verbose output."
        )
    ] = False
):
    '''
    Get the measured current from ACCURATE. \n
    To exit, press Ctrl+C. \n
    PORT accepts also the environment variable SERIAL_PORT.
    '''

    # Initialize variables
    femto_current_avg = 0
    count = 0
    # Initialize the console
    console = Console()
    # Initialize and write header to file
    if log is not None:
        with open(log, "w") as f:
            f.write("Timestamp, Instantaneous current (fA), Average current (fA)\n")
            f.write("Start time: " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + "\n")
            f.write("-----------------------------\n")

    with serial.Serial(port, baudrate, timeout=1) as ser:
        try:
            with Live(console=console, refresh_per_second=4) as live:
                while True:
                    # Check header
                    header = ser.read()
                    if header[0] != 0xDD:
                        continue

                    # Read data
                    ser_data = ser.read(6)
                    # Extract data
                    data = int.from_bytes(ser_data, byteorder='little')

                    # Instantaneous current
                    charge = data * lsb
                    atto_current = charge / (period * 1e-3)
                    femto_current = atto_current * 1e-3
                    # Average current
                    femto_current_avg += femto_current
                    count += 1
                    # Format current
                    scale_factor, unit = format_current(femto_current)

                    # Log to file in CSV format
                    if log is not None:
                        with open(log, "a") as f:
                            timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                            f.write(f"{timestamp}, {femto_current:.2f}, {femto_current_avg/count:.2f}\n")

                    # Create a text object with the current values
                    text = Text()
                    text.append("Live data:", style="bold")
                    text.append(f"\nInstantaneous current: {femto_current*scale_factor:.2f} {unit}")
                    text.append(f" - Average current: {(femto_current_avg*scale_factor/count):.2f} {unit}")
                    if verbose:
                        text.append(f"\n\nDebug Data:", style="bold dim")
                        text.append(f"\nSerial data: 0x{ser_data.hex()} - ", style="dim")
                        text.append(f"Integer representation: {data}", style="dim")
                    # Update the live display with the new text
                    live.update(text)

        except KeyboardInterrupt:
            # Ctrl+C pressed, exit
            raise typer.Exit()
        

@app.command()
def set_dac(
    port: Annotated[
        str, typer.Argument(
            help="Serial port where the FPGA is connected.",
            envvar="SERIAL_PORT"
        )
    ],
    channel: Annotated[
        str, typer.Argument(
            help="DAC channel to set [A-H]."
        )
    ],
    voltage: Annotated[
        float, typer.Argument(
            help="Voltage to set in volts."
        )
    ],
    baudrate: Annotated[
        int, typer.Option(
            "-b", "--baudrate",
            help="Baudrate for the serial communication."
        )
    ] = 9600,
):
    '''
    Set the DAC's channel voltages.
    '''
    # Convert voltage to DAC Din value
    Din = (voltage * 4096) // 3 # Vref = 3V, 12-bit DAC
    Din_binary = bin(int(Din))[2:].zfill(32)

    # Send data to FPGA
    with serial.Serial(port, baudrate, timeout=1) as ser:
        # Send address
        ser.write(DAC_address[channel].to_bytes())
        # Send data
        ser.write(Din_binary)
    
    print(f"Channel {channel} set to {voltage}V - ({Din_binary})")
    raise typer.Exit()




###################
# UTILITY FUNCTIONS
###################

def format_current(femto_current):
    """
    Formats the given femto current, by picking a more appropriate unit.

    Args:
        femto_current (float): The femto current to be formatted.

    Returns:
        scale_factor (float): The scale factor to be applied to the current.
        unit (str): The unit of the current.
    """
    
    if femto_current < 1e3:
        return 1, "fA"
    elif femto_current < 1e6:
        return 1e-3, "pA"
    elif femto_current < 1e9:
        return 1e-6, "nA"
    else:
        return 1e-9, "uA"
    
